/// $Id: mysqldb.C,v 1.13 2003/03/29 19:03:30 dvermeir Exp $ $Name:  $
#include <iostream>
#include <sstream>
#include <mysql.h>

#include <dvutil/ntostr.h>
#include "mysqldb.h"

namespace Dv {
namespace MySql {

using namespace std;
using Dv::Util::ntostr;
using Dv::Util::Props;

string
escape(const string& s) {
if (s.size()==0)
  return "";
else {
  char*		buf = new char[2*s.size()+1];
  unsigned int	len = mysql_escape_string(buf,s.data(),s.size());
  buf[len] = 0;
  string ss(buf,len);
  delete buf;
  return ss;
  }
}

/////////////////////// auxiliary static functions

static struct {
  int		type;
  const char*	name;
  } field_names[] = {
    {FIELD_TYPE_DECIMAL,	"DECIMAL"},
    {FIELD_TYPE_TINY,	"TINY"},
    {FIELD_TYPE_SHORT,	"SHORT"},
    {FIELD_TYPE_LONG,	"LONG"},
    {FIELD_TYPE_FLOAT,	"FLOAT"},
    {FIELD_TYPE_DOUBLE,	"DOUBLE"},
    {FIELD_TYPE_NULL,	"NULL"},
    {FIELD_TYPE_TIMESTAMP,	"TIMESTAMP"},
    {FIELD_TYPE_LONGLONG,	"LONGLONG"},
    {FIELD_TYPE_INT24,	"INT24"},
    {FIELD_TYPE_DATE,	"DATE"},
    {FIELD_TYPE_TIME,	"TIME"},
    {FIELD_TYPE_DATETIME,	"DATETIME"},
    {FIELD_TYPE_YEAR,	"YEAR"},
    {FIELD_TYPE_NEWDATE,	"NEWDATE"},
    {FIELD_TYPE_ENUM,	"ENUM"},
    {FIELD_TYPE_SET,	"SET"},
    {FIELD_TYPE_TINY_BLOB,	"TINY_BLOB"},
    {FIELD_TYPE_MEDIUM_BLOB,"MEDIUM_BLOB"},
    {FIELD_TYPE_LONG_BLOB,	"LONG_BLOB"},
    {FIELD_TYPE_BLOB,	"BLOB"},
    {FIELD_TYPE_VAR_STRING,	"VAR_STRING"},
    {FIELD_TYPE_STRING,	"STRING"},
    {0,			0}
    };

static const char*
field_type_name(unsigned int ft) {
int	t(ft);

for (unsigned int i=0;(field_names[i].name);++i)
  if (field_names[i].type==t)
    return field_names[i].name;
throw Sql::Exception(string("Unknown field type code: ")+ntostr(ft));
return 0;
}

static bool
field_type_isint(unsigned int ft) {
return	((ft==FIELD_TYPE_TINY) ||
   (ft==FIELD_TYPE_SHORT) ||
   (ft==FIELD_TYPE_LONG) ||
   (ft==FIELD_TYPE_INT24) ||
   (ft==FIELD_TYPE_LONGLONG) ||
   (ft==FIELD_TYPE_YEAR)); 
}

static bool
field_type_isflt(unsigned int ft) {
return	((ft==FIELD_TYPE_FLOAT) || (ft==FIELD_TYPE_DOUBLE));
}

static bool
field_type_isstr(unsigned int ft) {
return	(ft==FIELD_TYPE_STRING);
}

static bool
field_type_isbin(unsigned int ft) {
return	(ft==FIELD_TYPE_BLOB);
}

static void
fields_from_result(MYSQL_RES* result,vector<Sql::Field>& fields) {
fields.clear();
MYSQL_FIELD	*field;

while ((field=mysql_fetch_field(result))) {
  int	ft = field->type;
  fields.push_back(Sql::Field(
    field->name,				// name
    field->table,				// table
    field_type_name(ft),			// type
    "",					// default
    field_type_isint(ft),			// isint
    field_type_isflt(ft),			// isflt
    field_type_isstr(ft),			// isstr
    field_type_isbin(ft),			// isbin
    field->length));
  }
}

static void
get_vector_of_strings(MYSQL_RES* result,vector<string>& v) {
v.clear();

MYSQL_ROW	mrow;
while ((mrow = mysql_fetch_row(result))) {
  unsigned int ncols = mysql_num_fields(result);
  if (ncols!=1)
    throw Sql::Exception(  
            string("get_vector_of_strings ncols = ")+ ntostr(ncols)+" not 1");
  unsigned long* lengths = mysql_fetch_lengths(result);
  v.push_back(string(mrow[0],lengths[0]));
  }
mysql_free_result(result);
}

/////////////////////// Db


Db::Db(const char* dbname,const char* username,const char* passwd,const char* hostname):
  Sql::Db(dbname,hostname), mysql_(mysql_init(0)), user_(username?username:""),
  passwd_(passwd?passwd:"") {
const char*	h;

if (host()=="localhost")
  h = 0; // will use UNIX socket, *much* faster then tcp/ip
else
  h = host().c_str();

// note that we use parameter for username, dbname; *not* data members

if (!mysql_real_connect(static_cast<MYSQL*>(mysql_),h,username,passwd,dbname,0,0,0)) {
  ok(false);
  err(mysql_error(static_cast<MYSQL*>(mysql_)));
  }
}

// auxiliary functions
static const char* grab(const char* key,const Props& props,const string& prefix) {
const string* ps(props.find(prefix+key));
return (ps?ps->c_str():0);
}

static string grabs(const char* key,const Props& props,const string& prefix) {
const string* ps(props.find(prefix+key));
return (ps?*ps:string(""));
}

Db::Db(const Props& config,const string& prefix) throw (Sql::Exception):
  Sql::Db(grab("mysqldb",config,prefix),grab("mysqlhost",config,prefix)),
  mysql_(mysql_init(0)), user_(grabs("mysqluser",config,prefix)),
  passwd_(grabs("mysqlpasswd",config,prefix)) {

if (!mysql_real_connect(static_cast<MYSQL*>(mysql_),
       // will use UNIX socket on "localhost", *much* faster then tcp/ip
       ((host()=="localhost")?0:host().c_str()),
       user().c_str(),
       passwd().c_str(),
       database().c_str(),0,0,0)
    )
  throw Sql::Exception(string("cannot open db ")+database()+"@"+host());
}

Db::~Db() {
mysql_close(static_cast<MYSQL*>(mysql_));
}

Sql::Command_*
Db::command_(Sql::Command& cmd) {
return new Command_(*this,cmd);
}

bool
Db::database(const string& dbname) {
int error =  mysql_select_db(static_cast<MYSQL*>(mysql_),dbname.c_str());
if (error) {
  ok(false);
  err(mysql_error(static_cast<MYSQL*>(mysql_)));
  return false;
  }
name(dbname);
return ok();
}

vector<string>
Db::databases() {
MYSQL_RES*	result = mysql_list_dbs(static_cast<MYSQL*>(mysql_),0);

vector<string>	dbs;

if (!result) {
  ok(false);
  err(mysql_error(static_cast<MYSQL*>(mysql_)));
  return dbs;
  }

get_vector_of_strings(result,dbs);

return dbs;
}

vector<string>
Db::tables() {
MYSQL_RES*	result = mysql_list_tables(static_cast<MYSQL*>(mysql_),0);
vector<string>		tables;

if (!result) {
  ok(false);
  err(mysql_error(static_cast<MYSQL*>(mysql_)));
  return tables;
  }
get_vector_of_strings(result,tables);

return tables;
}

vector<Sql::Field>
Db::fields(const string& table) {
MYSQL_RES* result = mysql_list_fields(static_cast<MYSQL*>(mysql_),table.c_str(),0);
vector<Sql::Field>	fields;

if (!result) {
  ok(false);
  err(mysql_error(static_cast<MYSQL*>(mysql_)));
  return fields;
  }

fields_from_result(result,fields);
mysql_free_result(result);
return fields;
}


/////////////////////// Command_

Command_::Command_(Db& db, Sql::Command& cmd): Sql::Command_(cmd),
  db_(db), result_(0), row_(-1), mrow_(0), errno_(0) {
}

Command_::~Command_() {
if (result_)
  mysql_free_result(static_cast<MYSQL_RES*>(result_));
}

bool
Command_::check_error() {
static const char*	OK = "ok";

const char* errmsg(mysql_error(static_cast<MYSQL*>(db_.mysql_)));

if (errmsg&&(*errmsg!=0)) {
  ok(false);
  err(errmsg);
  return false;
  }
else	{
  ok(true);
  err(OK);
  return true;
  }
}

bool
Command_::exec() throw (Sql::Exception) {
if (cmd_.ran())
  return cmd_.ok();

run();

string	q(cmd_.text());

if (db_.log())
  *db_.log() << "dvmysql: " << q;

if (mysql_real_query(static_cast<MYSQL*>(db_.mysql_),q.data(),q.size())) {
  // error
  check_error();
  if (db_.log())
    *db_.log() << ": " << err() << endl;
  return false;
  }

result_	= mysql_store_result(static_cast<MYSQL*>(db_.mysql_));

if (result_) {
  nrows(mysql_num_rows(static_cast<MYSQL_RES*>(result_)));
  ncols(mysql_num_fields(static_cast<MYSQL_RES*>(result_)));

  if (db_.log())
    *db_.log() << " :" << nrows() << " rows" << endl;
  }
else
  if (mysql_field_count(static_cast<MYSQL*>(db_.mysql_))!=0) {
    // error
    check_error();
    if (db_.log())
    	*db_.log() << " :" << err() << endl;
    return false;
    }
  else {
    if (!check_error())
            throw Sql::Exception("Mysql::Command::exec() check_error() fails");
    if (db_.log())
    	*db_.log() << " :" << err() << endl;
    }
return true;
}

void
Command_::sync(unsigned int r) throw (Sql::Exception) {
if (!result_)
  throw Sql::Exception("Command_::sync(): 0 result_");

if ((int)r != row_) {
  row_ = r;
  mysql_data_seek(static_cast<MYSQL_RES*>(result_),row_);
  mrow_ = mysql_fetch_row(static_cast<MYSQL_RES*>(result_));
        if (mrow_==0)
      throw Sql::Exception("Command_::mrow_ == 0");
  lengths_ = mysql_fetch_lengths(static_cast<MYSQL_RES*>(result_));
  }
}

void
Command_::fetch(vector<string>& v,unsigned int r) throw (Sql::Exception) {
sync(r);


for (unsigned int c=0;(c<cmd_.ncols());++c)
  if (static_cast<MYSQL_ROW>(mrow_)[c])
    v.push_back(string(static_cast<MYSQL_ROW>(mrow_)[c],lengths_[c]));
  else
    v.push_back("NULL");
}

void
Command_::fetch(string& s,unsigned int r,unsigned int c) throw (Sql::Exception) { 
sync(r);
if (static_cast<MYSQL_ROW>(mrow_)[c])
  s =  string(static_cast<MYSQL_ROW>(mrow_)[c],lengths_[c]);
else
  s = "NULL";
}

const vector<Sql::Field>&
Command_::fields() {

if (!result_) {
  ok(false);
  err("no result yet");
  return fields_;
  }

if (fields_.size()==0)
  fields_from_result(static_cast<MYSQL_RES*>(result_),fields_);
return fields_;
}

void
Command_::escape(const string& s) {
ostringstream&	oss = textstream();
if (s.size()==0)
  oss << "NULL";
else {
  char*		buf = new char[2*s.size()+1];
  unsigned int	len = mysql_escape_string(buf,s.data(),s.size());
  oss << '\'';
  oss.write(buf,len);
  oss << '\'';
  delete buf;
  }
}

unsigned int
Command_::insertid() const {
return mysql_insert_id(static_cast<MYSQL*>(db_.mysql_));
}


}
}
