// $Id: mysqldb.h,v 1.11 2002/10/19 13:50:13 dvermeir Exp $ $Name:  $

#ifndef	MYSQLDB_H
#define	MYSQLDB_H


#include	<string>
#include	<vector>

#include	<dvmysql/sqldb.h>
#include	<dvutil/props.h>

/*! \file
    This file declares the mysql-dependent classes.

    <ul>
    <li> Dv::MySql::Db: an open database connection, derived from Dv::Sql::Db.
    <li> Dv::Sql::Command_: an implementation class derived from Dv::Sql::Command_.
    </ul>
*/

namespace Dv {

//! The Dv::MySql namespace contains Dv::Mysql::Db and Dv::Mysql::Command_
namespace MySql {

//! Return string that is acceptable as a field value in a mysql statement.
/*! This will insert appropriate quotes e.g. before an apostrophe etc.
    It will not add quotes around the string!
*/
std::string escape(const std::string&);

class Command_;

//! A connection to a mysql database.
class Db: public Sql::Db {
//! Command_ has access to mdbms-dependent private info, e.g. mysql_.
friend class Command_;
public:
  //! Open a connection, check result with Dv::Sql::Db::ok() and friends.
  Db(const char* dbname=0, const char* username="root", 
     const char* passwd=0,const char* hostname=0);
  //! Get database name etc. from configuration Dv::Util::Props object.
  /*! The configuration file uses the following attribute names
      (each preceded by prefix):
      \code
      mysqldb = 
      mysqluser = 
      mysqlpasswd = 
      mysqlhost = 
      \endcode
      If, e.g. prefix = "abc:", this would become
      \code
      abc:mysqldb = 
      abc:mysqluser = 
      abc:mysqlpasswd = 
      abc:mysqlhost = 
      \endcode
      Any missing or empty attribute is equivalent to a corresponding
      0 argument for the other constructor. 

      Example usage: (from tmysql.C in the distribution)
      \code
      try {
        static const std::string CONFIG("tmysql.config");
        ifstream	ifs(CONFIG.c_str());
        if (!ifs)
          throw runtime_error(CONFIG+": cannot open");
        
        Dv::Util::Props	config(ifs);
        std::string dictfn(config["dict"]);
  
        MySql::Db	db(config,"tmysql::");
        ..
      catch (exception& e) {
	cerr << e.what() << endl;
	return 1;
        }
      \endcode
      where ``tmysql.config'' may contain e.g.
      \code
      tmeta::mysqldb=mysql
      tmeta::mysqluser=root
      #tmeta::mysqlpasswd=
      #
      tmysql::mysqldb=test
      tmysql::mysqluser=root
      #mysqlpasswd=
      dict=/usr/dict/words
      \endcode
  */
  Db(const Dv::Util::Props& config,const std::string& prefix="") throw (Sql::Exception);
  //! Destructor.
  virtual ~Db();
  
  //! Return user of this database connection.
  std::string	user() const	{ return user_; }
  //! Return password used to open this connection.
  std::string	passwd() const	{ return passwd_; }
  //! Return name of current database of connection.
  std::string	database() const { return Sql::Db::database(); } 
  //! Switch connection to another database.
  bool		database(const std::string& database);
  //! Return the names of all databases available through this connection.
  std::vector<std::string>	databases();
  //! Return the names of all tables available in the current database.
  std::vector<std::string>	tables();
  //! Return ordered list of fields in a table.
  std::vector<Sql::Field>	fields(const std::string& table);


private:
  void*		mysql_;
  std::string	user_;
  std::string	passwd_;

  Db(const Db&);
  Db&		operator=(const Db&);

  //! Implementation of Sql::Db pure virtual factory function.
  Sql::Command_*	command_(Sql::Command& cmd);
};

//! Implementation class for Sql::Command helper.
class Command_: public Sql::Command_ {

//! Only Db can e.g. construct a Command_.
friend class Db;

private:
  Db&			db_;
  //! Stored result of query, opaque declaration, really MYSQL_RES*
  void*			result_;
  //! Current row# [0..] of mrow_, -1 if none.
  int			row_;	
  //! Lengths of fields [0..ncols()] in current row.
  unsigned long*	lengths_; 
  //! Buffer corresponding to row_; opaque declaration, really MYSQL_ROW.
  void*			mrow_;
  //! Error code, 0 if ok.
  int			errno_;
  //! Field list of the query result, filled on demand.
  std::vector<Sql::Field>	fields_;

  Command_(Db& db,Sql::Command& cmd);
  virtual ~Command_();

  //! Return true iff all is well.
  bool	check_error(); 
  //! Synchronize mrow_ cursor to point to row # r.
  void	sync(unsigned int r) throw (Sql::Exception);
  //! Return error code.
  int	error() const	{ return errno_; }
  //! Return MySql::Db connection of this command.
  Db&	db()		{ return db_; }

  //! Execute current command.
  /*! Part of Sql::Command_ required interface. */
  bool	exec() throw (Sql::Exception);
  //! Fetch r'th row from query result.
  /*! Part of Sql::Command_ required interface. */
  void	fetch(std::vector<std::string>&,unsigned int r) throw (Sql::Exception);
  //! Fetch c'th element in r'th row of query result.
  /*! Part of Sql::Command_ required interface. */
  void	fetch(std::string&,unsigned int r,unsigned int c) throw (Sql::Exception);
  //! Write quoted and escaped version of parameter to textstream().
  /*! Part of Sql::Command_ required interface. */
  void	escape(const std::string&);
  //! Return field description corresponding to query result.
  /*! Part of Sql::Command_ required interface. */
  const std::vector<Sql::Field>&	fields();
  /** Return auto_increment insertid of last query */
  unsigned int insertid() const;
};
}
}
#endif
