// $Id: sqldb.C,v 1.12 2003/08/17 12:58:22 dvermeir Exp $

#include <vector>
#include <algorithm>

#include <dvutil/ntostr.h>
#include <dvmysql/sqldb.h>

namespace Dv {
namespace Sql {

using namespace std;
using Dv::Util::ntostr;

const string Exception::NAME("Dv::Sql::Exception");
static const string ATTEMPT("attempt to append ");

///////////////////////	Field
Field::Field(const string& name,const string& table, const string& type,
  const string& defaultval,
  bool isint, bool isflt, bool isstr, bool isbin, unsigned int sz): 
    name_(name), table_(table), type_(type), default_(defaultval),
    int_(isint), float_(isflt),
    string_(isstr), binary_(isbin), size_(sz) {
}

ostream&
operator<<(ostream& os,const Field& f) {
os << f.type() << "[" << f.size() << "] " << f.table() << "." << f.name();
return os;
}

Util::Date 
Field::timestamp2date(const string& timestamp) {
Util::Date d;
d.year(Dv::Util::Int(string(timestamp,0,4),10));
d.month(Dv::Util::Int(string(timestamp,4,2),10)-1);
d.day(Dv::Util::Int(string(timestamp,6,2),10));
d.hours(Dv::Util::Int(string(timestamp,8,2),10));
d.minutes(Dv::Util::Int(string(timestamp,10,2),10));
d.seconds(Dv::Util::Int(string(timestamp,12,2),10));
return d;
}

///////////////////////	Db
Db::Db(const char* name,const char* host): host_(host?host:"localhost"),
  ok_(true), err_("ok"), log_(0), name_(name?name:"") {
}

///////////////////////	Command

Command::Command(Db& db,const char* txt): 
  ran_(false), err_("ok"), ok_(db), text_(), frozen_(false), nrows_(0), ncols_(0),
  cmd_(db.command_(*this)) {
if (txt)
  text_ << txt << " ";
}

Command::~Command() {
delete cmd_;
}


Command&
Command::operator<<(const char* s) throw (Exception) {
if (frozen_)
  throw Exception(ATTEMPT+"const char* \""+s+"\" to frozen query");

if (ran())
  return *this;

text_ << s << " ";
return *this;
}

Command&
Command::operator<<(const string& s) throw (Exception) {
if (frozen_)
  throw Exception(ATTEMPT+"string \""+s+"\" to frozen query");
if (ran())
  return *this;

cmd_->escape(s);
text_ << " ";
return *this;
}

Command&
Command::operator<<(int i) throw (Exception) {
if (frozen_)
  throw Exception(ATTEMPT+"int "+ntostr(i)+" to frozen query");
if (ran())
  return *this;
text_ << i << " ";
return *this;
}

Command&
Command::operator<<(unsigned int i) throw (Exception) {
if (frozen_)
  throw Exception(ATTEMPT+"unsigned int "+ntostr(i)+" to frozen query");
if (ran())
  return *this;
text_ << i << " ";
return *this;
}

Command&
Command::operator<<(short i) throw (Exception) {
if (frozen_)
  throw Exception(ATTEMPT+"short "+ntostr(i)+" to frozen query");
if (ran())
  return *this;
text_ << i << " ";
return *this;
}

Command&
Command::operator<<(unsigned short i) throw (Exception) {
if (frozen_)
  throw Exception(ATTEMPT+"unsigned short "+ntostr(i)+" to frozen query");
if (ran())
  return *this;
text_ << i << " ";
return *this;
}

Command&
Command::operator<<(long l) throw (Exception) {
if (frozen_)
  throw Exception(ATTEMPT+"long "+ntostr(l)+" to frozen query");
if (ran())
  return *this;
text_ << l << " ";
return *this;
}

Command&
Command::operator<<(unsigned long l) throw (Exception) {
if (frozen_)
  throw Exception(ATTEMPT+"unsigned long "+ntostr(l)+" to frozen query");
if (ran())
  return *this;
text_ << l << " ";
return *this;
}

Command&
Command::operator<<(float f) throw (Exception) {
if (frozen_)
  throw Exception(ATTEMPT+"float "+ntostr(f)+" to frozen query");
if (ran())
  return *this;
text_ << f << " ";
return *this;
}

Command&
Command::operator<<(double d) throw (Exception) {
if (frozen_)
  throw Exception(ATTEMPT+"double "+ntostr(d)+" to frozen query");
if (ran())
  return *this;
text_ << d << " ";
return *this;
}

bool
Command::valid(int r) const {
return (ran() && ok() && (r>=0) && ((unsigned int)r<nrows_));
}

Command::iterator
Command::begin() {
if (valid(0))
  return iterator(*this,iterator::BEGIN);
else
  return iterator(*this,iterator::END);
}


///////////////////////	Command_

// a default (and inefficient) implementation

void
Command_::fetch(string& s,unsigned int r,unsigned int c) throw (Exception) {
vector<string>	v;

fetch(v,r);
if ((c<0)||(c>=cmd_.ncols()))
  throw Exception(string("attempt to fetch column ")+ntostr(c)+
  " from results which is <0 or >max "+ntostr(cmd_.ncols()-1));
s = v[c];
}

void
Command_::escape(const string& s) {
ostringstream&	txt = textstream();

txt << '\'';

unsigned int n = s.size();

for (unsigned int i=0;(i<n);++i)
  {
  if ((s[i]=='\'')||(s[i]=='\\'))
    txt << '\\';
  txt << s[i];
  }
txt << "\' ";
}

unsigned int
Command_::insertid() const { // default implementation
return 0;
}

///////////////////////	Command::iterator

Command::iterator&
Command::iterator::operator++() throw (Exception) {
if (row_==END)
  return *this;
row_++;
if (row_<0)
  throw Exception("Command::iterator::operator++ row_<0");
if (!cmd_.valid(row_))
  row_ = END;
return *this;
}

vector<string>
Command::iterator::operator*() const throw (Exception) {
vector<string>	v; // (cmd_.ncols());

if (row_<0)
  throw Exception("Command::iterator::operator* row_<0");

cmd_.fetch(v,(unsigned int)row_);
return v;
}

string
Command::iterator::operator()(unsigned int c) const throw (Exception) {
if ((c<0)||(c>=cmd_.ncols()))
  throw Exception(string("Command::iterator::operator(")+ntostr(c)+
  ") with parameter <0 or >max "+ntostr(cmd_.ncols()-1));
string	s;

cmd_.fetch(s,row_,c);
return s;
}
}
}

using namespace std;
using namespace Dv;

Util::Props& operator<<(Util::Props& props, const Sql::Command::iterator& i) {
Sql::Command& cmd(i.command());
const vector<Sql::Field>& fields(cmd.fields());
for (size_t j=0; j<cmd.ncols(); ++j)
  props.add(fields[j].name(), i(j));
return props;
}
