// $Id: sqldb.h,v 1.16 2003/02/13 19:36:09 dvermeir Exp $ $Name:  $

#ifndef	DV_SQLDB_H
#define	DV_SQLDB_H

#include <string>
#include <sstream>
#include <iostream>
#include <vector>
#include <stdexcept>
#include <iterator>

#include <dvutil/date.h>
#include <dvutil/props.h>
#include <dvmysql/vector_string_input_iterator.h>

/*! \mainpage
\author dvermeir@vub.ac.be
 
<a href="../download/">Download</a>

\section installation Installation
 
Standard. See the INSTALL file in the top directory.

Dependencies:
	<ul>
	<li> <a href="http://tinf2.vub.ac.be/~dvermeir/software/dv/index.html">dvutil</a>
	<li> <a href="http://mysql.com/">mysql </a>
	</ul>
*/

/*! \file
    This file describes the dbms-independent classes:

    <ul>
    <li> Dv::Sql::Exception: thrown upon failure of some operations.
    <li> Dv::Sql::Field: describes a field in a table or query result.
    <li> Dv::Sql::Db: an open database connection.
    <li> Dv::Sql::Command: an sql command (query, update).
    <li> Dv::Sql::Command::iterator: an iterator for the result of a
         query.
    <li> Dv::Sql::Command_: an implementation class that must be subclassed
         for each particular DBMS.
    </ul>
*/

//! The whole package lives in this namespace.
namespace Dv {

//! Namespace for dbms-independent classes.
namespace Sql {

//! Exception class used throughout the package.
/*!
Derived from runtime_error, i.e. the only useful member function
is Dv::Sql::Exception::what() which returns constant C string
describing the error.

Most exceptions thrown are really logical errors (i.e. errors
in the application or package source code).
*/
class Exception: public std::runtime_error {
public:
  //! Name of class, is prepended to each message argument of the constructor.
 static const std::string NAME;
  //! Constructor, prepends NAME to message to obtain runtime_error::what().
 Exception(const std::string& message): std::runtime_error(NAME+": "+message) {}
};

//! A field/column in a table or a query result.
class Field {
public:
  //! Field constructor. Users will never use this.
  /*! What are valid arguments is dbms-dependent.
  */
  Field(const std::string& name,const std::string& table, const std::string& type,
  	const std::string& defaultvalue,
  	bool isint,bool isflt, bool isstr, bool isbin, unsigned int sz);

  //! The name of the field.
  std::string	name() const 	{ return name_; }
  //! The name of the table containing this column.
  /*! For a query result (i.e. a Dv::Sql::Command), the table
      can be defined by the ``as'' clause in a ``select'' statement.
  */
  std::string	table() const 	{ return table_; }
  //! The type of this field. The return value is dbms-dependent.
  std::string	type() const	{ return type_; }
  //! The default value (see the constructor).
  std::string	defaultvalue() const	{ return default_; }

  //! Does the field have a numeric (int or float) type?
  bool		numeric() const	{ return int_ || float_ ; }
  //! Does the field have an integer type?
  bool		integer() const	{ return int_; }
  //! Does the field have floating point type?
  bool		real() const	{ return float_; }
  //! Does the field have a binary type?
  bool		binary() const	{ return binary_; }
  //! Does the field have a string type?
  bool		text() const	{ return string_; }

  //! Convert mysql timestamp(14) value, e.g. 20010724110830 to Date.
  static Dv::Util::Date timestamp2date(const std::string& timestamp);
  //! Size of the field (in bytes).
  unsigned int	size() const	{ return size_; }

  //! Print readable representation on ostream.
  /*! Example:
      \code
       LONG[11] test.number
       STRING[32] test.word
       DATETIME[19] structuur2.wanneer
       VAR_STRING[100] structuur2.info
       TIMESTAMP[14] structuur2.time_stamp        
      \endcode
  */
  friend std::ostream& operator<<(std::ostream&,const Field&);
private:
  std::string	name_;
  std::string	table_;
  std::string	type_;
  std::string	default_;

  bool		int_;
  bool		float_;
  bool		string_;
  bool		binary_;
  unsigned int	size_;

};

class Command_;
class Command;

//! An connection to an SQL database.
class Db {

//! Friend.
friend class Command;

public:
  //! Since this class will be subclassed, its destructor must be virtual.
  /*! The destructor will close the connection.
  */
  virtual ~Db() {}

  //! Return true iff status is ok; e.g. db was succesfully opened.
  bool		ok() const		{ return ok_; }
  //! Same as Dv::Sql::Db::ok().
  operator	bool() const		{ return ok_; }
  //! Same as !Dv::Sql::Db::ok().
  bool		operator!() const	{ return !ok_; }

  //! Return string representation of last error.
  std::string	err() const		{ return err_; }
  //! Name of host machine on which database resides.
  std::string	host() const		{ return host_; }
  //! Name of current database of connection.
  std::string	database() const	{ return name_; }

  //! Logging is turned on if log()!=0.
  /*! The amount of logging is decided by the DBMS-dependent classes.
      E.g. Dv::MySql::Db logs the execution of every command as
      well as indication of the result.
  */
  std::ostream*	log() const		{ return log_; }
  //! Turn logging on/off, see Db::log().
  void		log(std::ostream* os) 	{ log_ = os; }

  //! Switch connection to database, return true iff succeeded.
  virtual bool		 database(const std::string& database) = 0;
  //! Return names of available databases.
  virtual std::vector<std::string> databases() = 0;
  //! Return names of tables in databases.
  virtual std::vector<std::string> tables() = 0;
  //! Return fields in table.
  virtual std::vector<Field> fields(const std::string& table) = 0;

protected:
  //! Set status.
  bool		ok(bool b)			{ return ok_ = b; }
  //! Set error string returned by Dv::Sql::err().
  void		err(const std::string& s)		{ err_ = s; }
  //! Set database name.
  void		name(const std::string& dbname)	{ name_ = dbname; }


  //! Return dbms-specific command.
  /*! Factory method; a concrete subclass X of Db implements this
      to return (a pointer to) an instance of the subclass of Command_
      corresponding to X's Command_.
  */
  virtual Command_*	command_(Command&) = 0;

  //! Constructor.
  /*! The constructor takes the name of the database
      and the name of the host where the database resides.
      If name == 0, there is no database yet, if host == 0,
      we use "localhost"

      Copy ctor and assignment are forbidden.
  */
  Db(const char* name,const char* host);

private:
  std::string	host_;
  bool		ok_;
  std::string	err_;
  std::ostream*	log_;

  //! Name of database, may be "".
  std::string	name_; 

  //! Cctor is forbidden.
  Db(const Db&);
  //! Assignment is forbidden.
  Db&		operator=(const Db&);

};


//! A Command represents a command, e.g. a query, for a Dv::Sql::Db.
/*! Command is a dbms-independent proxy class for the dbms-dependent
    Command_. A Command object has a pointer to a companion Command_
    object, which has been created using Dv::Sql::Db::command_(). 

    This is an alternative for inheritance:
    instead of subclassing Command, the ``subclass part'' is
    available in *cmd_.

    This has the advantage that only the definition of the
    database object ``db'' derived from Dv::Sql::Db depends on the actual
    DBMS used. All other database manipulatsions can be done
    through normal Dv::Sql::Command objects, which are initialized
    using a referrence to a Dv::Sql::Db. Consequently, applications
    are more independent of the actual underlying DBMS.

    Example:
    \code
    MySql::Db db(...);

    std::string name;
    Sql::Command query("select * from person where name = ");
    query << name;
    if (q.exec())
      for (Sql::Command::iterator i=q.begin(); i!=q.end(); ++i) 
        cout << i(0) << "\t" << i(1) << endl;
    \endcode
    Commands contain a text. This text is built using
    Command::operator<<(..).
*/
class Command {

public:

  class iterator;
  //! Friend.
  friend class Db;
  //! Friend.
  friend class iterator;
  //! Friend.
  friend class Command_;


  //! Constructor, text is initial text of command.
  /*! Example
      \code
      Sql::Command	q("insert into person values (");
      // rest of command is inserted using operator<<
      \endcode
  */
  Command(Db& db,const char* text=0);
  //! Destructor also destroys any result data.
  ~Command();

  //! Status of command.
  bool		ok() const	{ return ok_; }
  //! Same as Dv::Sql::Command::ok().
  operator	bool() const	{ return ok_; }
  //! Error message.
  std::string	err() const	{ return err_; }
  //! Return true iff the command has already been executed.
  /*! Calling Dv::Sql::Command::exec() twice is a noop.
  */
  bool		ran() const	{ return ran_; }
  //! Return database connection with which this command is associated.
  Db&		db(); // see below, for inline definition
  //! Return text of command.
  /*! Example
    \code
    std::string name("jane");
    Sql::Command query("select * from person where name = ");
    query << name;
    cout << query.text();
    \endcode
    will print
    \code
    select * from person where name = 'jane'
    \endcode
  */
  std::string	text() 	{ frozen_ = true; return text_.str();}
  						    
  //! Return field information for query result.
  const std::vector<Field>&	fields();
  //! Return number of rows in query result.
  unsigned int	nrows()		{ return nrows_; }
  //! Return number of columns in query result.
  unsigned int	ncols()		{ return ncols_; }
  //! Return logical pointer to first row of the result.
  iterator	begin();
  //! Return logical pointer to past the end of the query result.
  iterator	end();

  //! Check whether a rowno r is valid, i.e. the r'th row is available
  bool		valid(int r) const;

  //! Append a string to the command text; the string will be escaped and quoted.
  /*! 
      E.g. after
      \code
      Db::Sql::Db& db;
  
      Dv::Sql::Command	q(db,"select * from person");
      q << "where name ="  << std::string("joe's nightamere");
      \endcode
      q.text_ will contain
      \code
      select * from person where name = 'joe\'s nightmare'
      \endcode

      Note that Dv::Sql::Command::operator<<(const char*) does not escape or quote.
  */
  Command&	operator<<(const std::string&) throw (Exception);
  //! Append a string to the command text.
  Command&	operator<<(const char*) throw (Exception);
  //! Append an int to the command text.
  Command&	operator<<(int) throw (Exception);
  //! Append an unsigned int to the command text.
  Command&	operator<<(unsigned int) throw (Exception);
  //! Append a short to the command text.
  Command&	operator<<(short) throw (Exception);
  //! Append an unsigned short to the command text.
  Command&	operator<<(unsigned short) throw (Exception);
  //! Append a long to the command text.
  Command&	operator<<(long) throw (Exception);
  //! Append an unsigned long to the command text.
  Command&	operator<<(unsigned long) throw (Exception);
  //! Append a float to the command text.
  Command&	operator<<(float) throw (Exception);
  //! Append a double to the command text.
  Command&	operator<<(double) throw (Exception);

  //! Execute the query, status is determined by underlying DBMS.
  bool		exec(); // inline definition at the end of this file
  /** Return auto_increment id of last insert. */
  unsigned int insertid() const;
private:
  bool		ran_;
  std::string	err_;
  bool		ok_;
  std::ostringstream text_;
  bool		frozen_; // after text_.str() has been used


  //! Nrows_ is # of rows in query answer, only valid if ran() && ok()
  unsigned long	nrows_;
  //! Ncols_ is # of columns in query answer, only valid if ran() && ok()
  unsigned long	ncols_;
  

  //! Associated dbms-dependent object.
  /*! Many functions such as Dv::Sql::Command::exec() are delegated
      to cmd_->exec().
  */
  Command_	*cmd_;

  //! Retrieve the r'th row of the result ( 0<= r < nrows_ ).
  /*! Delegated to cmd_. */
  void 	fetch(std::vector<std::string>& v,unsigned int r) throw (Exception);
  //! Retrieve the r'th row and c'th column of the result ( 0<= r < //nrows_ ).
  /*! Delegated to cmd_. */
  void 	fetch(std::string& s,unsigned int r,unsigned int c) throw (Exception);


  //! Cctor is forbidden.
  Command(const Command&); 
  //! Assignment is forbidden.
  Command&	operator=(const Command&);

  void		nrows(unsigned int n)	{ nrows_ = n; }
  void		ncols(unsigned int n)	{ ncols_ = n; }
  void		run()			{ ran_ = true; }
  void		err(const char*s)	{ err_ = std::string(s); }
  void		ok(bool b)		{ ok_ = b; }

};

//! A logical pointer (input iterator) inside the query result of a Dv::Sql::Command.
/*! A valid iterator in the range [begin,end[ points to
    a row of the query result.

    Its value type is the type of a row in the query result, i.e.
    a vector of strings (classes like Dv::Int can be used to
    convert to numeric types, see dvutil).
*/
class Command::iterator: public vector_string_input_iterator {

//! Friend.
friend class Command;

public:
  //! Make the iterator point to the next row in the result: prefix version.
  iterator&	operator++() throw (Exception);
  //! Make the iterator point to the next row in the result: postfix version.
  iterator	operator++(int) { iterator tmp=*this; ++*this; return tmp; }

  //! Dereference to the current row.
  /*! NULL is represented by the std::string ``NULL''.
  */
  std::vector<std::string>	operator*() const throw (Exception);

  //! Compare two iterators.
  bool	operator==(const iterator&j) const {
  			return row_ == j.row_ && &cmd_ == &j.cmd_; }
  //! Compare two iterators.
  bool	operator!=(const iterator&j) const {
  			return row_ != j.row_ || &cmd_ != &j.cmd_; }

  //! Non standard operator: get c'th column of current row.
  /*! Example
      \code
      MySql::Db db(...);
  
      std::string name;
      Sql::Command query("select * from person where name = ");
      query << name;
      if (q.exec())
        for (Sql::Command::iterator i=q.begin(); i!=q.end(); ++i) 
          cout << i(0) << "\t" << i(1) << endl;
      \endcode
  */
  std::string operator()(unsigned int c) const throw (Exception);

  //! Return command associated with iterator.
  Command& command() const { return cmd_; }
private:
  enum { END = -1, BEGIN = 0 };

  Command&	cmd_;
  int		row_;	// may be BEGIN || END || 0<= row_ < cmd_.nrows_;

  iterator(Command& cmd,int r): cmd_(cmd), row_(r) {}
};

//! Proxy class for Sql::Db::Command. Not used by end users.
/*! This class needs to be subclassed for a particular DBMS.
    A Dv::Sql::Command_ object forms a pair with a Dv::Sql::Command
    object: each contains a reference (or pointer) to the other.
    The Dv::Sql::Command_ object associated with a Dv::Sql::Command
    object cmd is created using the Dv::Sql::Db virtual member 
    function Sql::Db::command_(cmd).
    
    There is delegation in both ways: Dv::Sql::Command_ uses 
    Dv::Sql::Command to perform some DBMS-independent operations and
    Dv::Sql::Command delegates all DBMS-dependent
    operations to Dv::Sql::Command_.
*/
class Command_ {
//! Friend.
friend class Command;
protected:
  //! Return number of rows in the query result.
  unsigned int	nrows()			{ return cmd_.nrows(); }
  //! Set number of rows in the query result.
  void		nrows(unsigned int n)	{ cmd_.nrows(n); }
  //! Set number of columns in the query result.
  void		ncols(unsigned int n)	{ cmd_.ncols(n); }
  //! Mark the command (and associated Dv::Sql::Command) as having been executed.
  void		run()			{ cmd_.run(); }
  //! Return status string of this command.
  std::string	err()			{ return cmd_.err(); }
  //! Set status string of this command.
  void		err(const char*s)	{ cmd_.err(s); }
  //! Set status of this command.
  void		ok(bool b)		{ cmd_.ok(b); }
  //! Retrieve text of the command as a string stream.
  std::ostringstream&	textstream()		{ return cmd_.text_; }

  //! Associated DBMS-independent Dv::Sql::Command object.
  Command&	cmd_;

  //! Constructor: the associated Dv::Sql::Command object must already exist.
  Command_(Command& cmd): cmd_(cmd) {}
  //! Destructor, virtual because it will be subclassed.
  virtual ~Command_() {}
  //! Actually execute the command, DBMS-dependent.
  virtual bool	exec() = 0;
  //! Return database connection associated with this command.
  virtual Db&	db() = 0;

  //! Write s in properly quoted and escaped form to this->textstream()
  virtual void	escape(const std::string& s);
  //! Low level retrieval: put row # r from query result into the parameter.
  virtual void	fetch(std::vector<std::string>&,unsigned int r) throw (Exception) = 0;
  //! Low level retrieval: get column c from row # r from query result.
  /*! An exception is thrown if any of r or c is out of range.
  */
  virtual void	fetch(std::string&,unsigned int r,unsigned int c) throw (Exception);
  //! Return field information for the query result.
  virtual const std::vector<Field>&	fields() = 0;
  /** Return id of auto_increment field of last insertion */
  virtual unsigned int insertid() const;

private:
  Command_(const Command_&);
  Command_& operator=(const Command_&);

};

inline void
Command::fetch(std::vector<std::string>& v,unsigned int r) throw (Exception) {
cmd_->fetch(v,r);
}

inline	void
Command::fetch(std::string& s,unsigned int r,unsigned int c) throw (Exception) {
cmd_->fetch(s,r,c);
}

inline bool Command::exec() { return cmd_->exec(); }
inline const std::vector<Field>& Command::fields() { return cmd_->fields(); }
inline Db& Command::db() { return cmd_->db(); }
inline Command::iterator Command::end() { return iterator(*this,iterator::END); }
inline unsigned int Command::insertid() const { return cmd_->insertid(); }

}
}

//! Update props to reflect query result row represented by iterator.
/*! Updat props such that props[fieldname] = result for each fieldname
    of the command associated with the iterator.
*/
Dv::Util::Props& operator<<(Dv::Util::Props&, const Dv::Sql::Command::iterator&);

#endif
